""" Non-Deterministic Buchi Automaton (NBA) and Generalized Non-Deterministic Buchi Automaton (GNBA)

A NBA is the tuple: <states, alphabet, tr, init, final>

A state is an integer number of a tuple of states (inductive definition)
The alphabet is a finite set of symbols
A transition relation tr is a tuple in State x Action x State
init is a subset of states
final is a subset of states

Example of use:

# Creates the states
s1 = (1,)
s2 = (2,)

# Creates the NBA that accept the language "a (b)^omega"
t = NBA(set([s1, s2]),
        set(["a","b"]),
        set([(s1,"a",s2),(s2,"b",s2)]),
        set([s1]),
        set([s2]))
"""

from __future__ import annotations
from collections.abc import Mapping, Iterator, Iterable

from typing import (
    TypeAlias, Type, Tuple, Callable, Optional, NewType,
    List, Set,
    Union, TypedDict, IO, TypeVar
)

from mc308.data_structures.ts import States, Prop, Trans
from mc308.data_structures.formula import PropFormula

from functools import reduce, partial
import itertools



SymbolicTrans = Tuple[States, PropFormula, States]
ExtendedTrans = Trans | SymbolicTrans


T = TypeVar('T')
class NBA():

    def __init__(self : NBA,
                 states : frozenset[States],
                 alphabet : frozenset[T],
                 tr : frozenset[ExtendedTrans],
                 init : frozenset[States],
                 final : frozenset[States]) -> None:
        self.states = set([a for a in states])
        self.alphabet = set([a for a in alphabet])
        self.tr = set([t for t in tr])
        self.init = set([i for i in init])
        self.final = set([i for i in final])

        self.nextmap = {s : set() for s in self.states}
        for (src, a, dst) in self.tr:
            self.nextmap[src].add((a,dst))


class GNBA(NBA):
    def __init__(self : GNBA,
                 states : frozenset[States],
                 alphabet : frozenset[T],
                 tr : frozenset[ExtendedTrans],
                 init : frozenset[States],
                 final : frozenset[set[States]]) -> None:
        self.states = set([a for a in states])
        self.alphabet = set([a for a in alphabet])
        self.tr = set([t for t in tr])
        self.init = set([i for i in init])
        self.final = set([frozenset([j for j in i]) for i in final])

        self.nextmap = {s : set() for s in self.states}
        for (src, a, dst) in self.tr:
            self.nextmap[src].add((a,dst))


    def to_nba(self: GNBA) -> NBA:
        """
        Degeneralization procedure.
        """



        raise NotImplementedError("GNBA to NBA construction not implemented")


